/*
 * Copyright (C) 2022 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */

package wrest

import (
	"database/sql"
	"fmt"
	"gitlab.nrao.edu/ssa/gocapo/capo/config"
	"gitlab.nrao.edu/ssa/gocapo/helpers"
	"os"
	"strconv"
	"strings"
)

// databaseType represents the database driver we are using (Postgres or MySQL
// for now)
type databaseType string

const (
	Postgres databaseType = "postgres"
	MySQL    databaseType = "mysql"
)

// defaultPort returns the default server port for the given database type. This
// is needed to connect
func (databaseType databaseType) defaultPort() int {
	if databaseType == Postgres {
		return 5432
	} else {
		return 3306
	}
}

// jdbcPrefix return the JDBC string prefix that will sadly be prepended to the
// URL in the Capo profile
func (databaseType databaseType) jdbcPrefix() string {
	if databaseType == Postgres {
		return "jdbc:postgresql://"
	} else {
		return "jdbc:mysql://"
	}
}

// connectionString formats the connection string for this database flavor, which
// we can then hand to the sql package.
func (flavor databaseFlavor) connectionString(host string, port int, user string, password string, dbName string) string {
	if flavor.Type == Postgres {
		return fmt.Sprintf("host=%s port=%d user=%s password=%s dbname=%s sslmode=disable", host, port, user, password, dbName)
	} else {
		return fmt.Sprintf("%s:%s@tcp(%s:%d)/%s", user, password, host, port, dbName)
	}
}

// databaseFlavor represents the information we need to build a connection to
// some database, based on what we know a priori and from Capo
type databaseFlavor struct {
	CapoPrefix      string
	UserCapoKey     string
	PasswordCapoKey string
	UrlCapoKey      string
	Type            databaseType
}

// archiveDB holds the information necessary to connect to the archive database
var archiveDB = databaseFlavor{
	CapoPrefix:      "edu.nrao.ssa.project.db",
	UserCapoKey:     "username",
	PasswordCapoKey: "password",
	UrlCapoKey:      "url",
	Type:            Postgres,
}

// proposalsDB holds the information necessary to connect to the proposals database
var proposalsDB = databaseFlavor{
	CapoPrefix:      "my.nrao.jdbc",
	UserCapoKey:     "username",
	PasswordCapoKey: "password",
	UrlCapoKey:      "URL",
	Type:            MySQL,
}

// getConnection uses the information in a particular database flavor to connect
// to a certain database
func getConnection(flavor databaseFlavor) (*sql.DB, error) {
	// Get db info and build string to get connection
	prop, err := config.InitConfig(os.Getenv("CAPO_PROFILE"), helpers.DefaultCapoPath)
	if err != nil {
		return nil, fmt.Errorf("unable to load CAPO config: %v", err)
	}

	// obtain the Capo settings for this flavor
	settings := prop.SettingsForPrefix(flavor.CapoPrefix)

	// get the user, password, and URL settings
	user := settings.GetString(flavor.UserCapoKey)
	password := settings.GetString(flavor.PasswordCapoKey)
	url := settings.GetString(flavor.UrlCapoKey)

	// remove the JDBC prefix
	url = strings.TrimPrefix(url, flavor.Type.jdbcPrefix())

	// split the URL into host/database sections
	splitUrl := strings.Split(url, "/")
	if len(splitUrl) != 2 {
		return nil, fmt.Errorf("URL does not contain 'host/database' pattern: %s", url)
	}
	host := splitUrl[0]
	dbName := splitUrl[1]

	// split the host into host and port
	splitHostPort := strings.Split(host, ":")
	host = splitHostPort[0]

	// use the default port if there is no port component, otherwise use the port component
	port := flavor.Type.defaultPort()
	if len(splitHostPort) > 1 {
		port, err = strconv.Atoi(splitHostPort[1])
		if err != nil {
			return nil, fmt.Errorf("unable to convert %s to integer: %v", splitHostPort[1], err)
		}
	}

	// build the connect string
	connInfo := flavor.connectionString(host, port, user, password, dbName)

	// open the database connection and check for errors
	db, err := sql.Open(string(flavor.Type), connInfo)
	if err != nil {
		return nil, fmt.Errorf("unable to open the database connection to %s@%s: %v", dbName, host, err)
	}

	// attempt to connect to the database and check for errors
	err = db.Ping()
	if err != nil {
		return nil, fmt.Errorf("unable to ping the database connection for %s@%s: %v", dbName, host, err)
	}

	// OK, we have a valid database connection
	return db, nil
}

// GetArchiveDBConnection returns a connection to the archive database
func GetArchiveDBConnection() (*sql.DB, error) {
	return getConnection(archiveDB)
}

// GetProposalsDBConnection returns a connection to the proposals database
func GetProposalsDBConnection() (*sql.DB, error) {
	return getConnection(proposalsDB)
}
